import boto3
import re

console = boto3.client('apigateway')


def fetch_api_by_name(api_name):
    """ Fetch an api record by its name """
    api_records = console.get_rest_apis()['items']
    matches = filter(lambda x: x['name'] == api_name, api_records)
    if not matches:
        return None
    return matches[0]


def fetch_resources_by_api(api_id):
    """ Fetch all resources for an apigateway rest api """
    return console.get_resources(restApiId=api_id)['items']


def fetch_resource(api_id, resource_id):
    """ Fetch extra metadata for this particular resource """
    return console.get_resource(restApiId=api_id, resourceId=resource_id)


def fetch_method(api_id, resource_id, verb):
    """ Fetch extra metadata for this particular method """
    return console.get_method(
            restApiId=api_id,
            resourceId=resource_id,
            httpMethod=verb)


def lambda_arn_from_method_integration_arn(int_arn):
    pattern = '.*functions/(.*)/invocations'
    matches = re.match(pattern, int_arn)
    return matches.group(1)


def lambda_name_from_lambda_arn(lambda_arn):
    return lambda_arn.split(':')[-1]


class RestApi(object):
    def __init__(self, name):
        self.name = name
        self.aws_id = None
        self.resources = dict()
        self._fetch_from_aws()
        if self.exists_in_aws:
            self._fetch_resources()

    def _fetch_from_aws(self):
        record = fetch_api_by_name(self.name)
        if record:
            self.aws_id = record['id']

    @property
    def exists_in_aws(self):
        return (self.aws_id is not None)

    def _fetch_resources(self):
        for record in fetch_resources_by_api(self.aws_id):
            path = record['path']
            if path not in self.resources:
                self.resources[path] = RestResource(path)
            resource = self.resources[path]
            resource.aws_id = record['id']
            resource.api = self
            resource.fetch_from_aws()


class RestResource(object):
    def __init__(self, path):
        self.path = path
        self.aws_id = None
        self.api = None
        self.methods = dict()

    @property
    def exists_in_aws(self):
        return (self.aws_id is not None and self.api is not None)

    def fetch_from_aws(self):
        if not self.exists_in_aws:
            raise Exception("Not enough info to query AWS for Rest Resource")
        record = fetch_resource(self.api.aws_id, self.aws_id)
        if 'resourceMethods' in record:
            for verb in record['resourceMethods']:
                if verb not in self.methods:
                    self.methods[verb] = RestMethod(verb)
                method = self.methods[verb]
                method.resource = self
                method.api = self.api
                method.fetch_from_aws()


class RestMethod(object):
    def __init__(self, verb):
        self.verb = verb
        self.resource = None
        self.api = None
        self.lambda_name = None
        self.lambda_function = None

    def fetch_from_aws(self):
        if (self.resource is None or self.api is None):
            msg = """I cannot possibly question AWS
            without knowing my resource and api ids"""
            raise Exception(msg)
        record = fetch_method(self.api.aws_id, self.resource.aws_id, self.verb)
        arn = record['methodIntegration']['uri']
        lambda_arn = lambda_arn_from_method_integration_arn(arn)
        self.lambda_name = lambda_name_from_lambda_arn(lambda_arn)



from psychic_disco.util import cp, zip, py_files_in_directory
from virtualenvapi.manage import VirtualEnvironment

import boto3
import os
import psychic_disco
array_zip = zip

_default_package = None


def default():
    global _default_package
    if _default_package is None:
        _default_package = DeploymentPackage()
    return _default_package


class DeploymentPackage(object):
    def __init__(self, key=None, bucket=None):
        conf = psychic_disco.config
        self.s3_key = key or conf.default_s3_key_for_deployment_package
        self.s3_bucket = bucket or conf.default_s3_bucket
        self.deployed = False

    @property
    def exists(self):
        return os.path.exists(self.localpath)

    @property
    def bucket(self):
        return self.s3_bucket

    @property
    def key(self):
        return self.s3_key

    @property
    def localpath(self):
        return os.path.join(self.bundle_dir, self.s3_key)

    @property
    def bundle_dir(self):
        return ".psychic_disco"

    def s3_path(self):
        return "s3://%s/%s" % (self.s3_bucket, self.s3_key)

    def deploy(self):
        if not self.deployed:
            s3 = boto3.client('s3')
            if not self.exists:
                self.bundle()
            with open(self.localpath, "r") as f:
                s3.put_object(Body=f, Bucket=self.s3_bucket, Key=self.s3_key)
                self.deployed = True

    def _install_packages_to_environment(self, env, requirements_file):
        with open(requirements_file, 'r') as f:
            for line in f:
                package_spec = line.rstrip()
                env.install(package_spec)

    def setup_virtualenv(self, venv_dir):
        os.environ['LC_ALL'] = 'C'
        # Contrary to docs, this won't actually initialize anything
        # until you either query or manipulate the installed packages
        env = VirtualEnvironment(venv_dir)
        requirements_file = "requirements.txt"
        if os.path.exists(requirements_file):
            self._install_packages_to_environment(env, requirements_file)
        return env.installed_packages

    def relative_path(self, path):
        return os.path.join(self.bundle_dir, path)

    @property
    def lambda_root(self):
        d = self.relative_path("lib/python2.7/dist-packages")
        if not os.path.exists(d):
            d = self.relative_path("lib/python2.7/site-packages")
        return d

    def bundle(self):
        self.setup_virtualenv(self.bundle_dir)
        for python_file in py_files_in_directory("."):
            cp(python_file, self.lambda_root)
        zip(self.lambda_root, self.localpath)



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


def requirements():
    with open('requirements.txt') as f:
        return [line.rstrip('\n') for line in f]


def version():
    with open('version.txt') as f:
        return f.read().rstrip('\n')

setup(name='psychic_disco',
      version=version(),
      description='Pythonic Microservices on AWS Lambda',
      long_description=readme(),
      url='http://github.com/robertdfrench/psychic-disco',
      author='Robert D. French',
      author_email='robert@robertdfrench.me',
      license='MIT',
      packages=['psychic_disco'],
      zip_safe=False,
      include_package_data=True,
      test_suite='nose.collector',
      tests_require=['nose'],
      install_requires=requirements(),
      scripts=['bin/psychic_disco'])
